import re

import path
import pytest

import tbump.git


@pytest.fixture()
def tmp_path(tmpdir):
    return path.Path(tmpdir)


@pytest.fixture
def test_path():
    this_dir = path.Path(__file__).parent
    return this_dir.abspath()


@pytest.fixture
def complex_version_regex():
    pattern = """\
  (?P<major>\d+)
  \.
  (?P<minor>\d+)
  \.
  (?P<patch>\d+)
  (
    -
    (?P<channel>alpha|beta)
    -
    (?P<release>\d+)
  )?
  """
    return re.compile(pattern, re.VERBOSE)


def assert_in_file(*args):
    parts = args[0:-1]
    first = parts[0]
    rest = parts[1:]
    last = parts[-1]
    expected_line = args[-1]
    file_path = path.Path(first).joinpath(*rest)
    for line in file_path.lines():
        if expected_line in line:
            return
    assert False, "No line found matching %s in %s" % (expected_line, last)


def setup_repo(tmp_path, test_path):
    src_path = tmp_path.joinpath("src")
    src_path.mkdir()
    test_path.joinpath("tbump.toml").copy(src_path)
    test_path.joinpath("VERSION").copy(src_path)
    test_path.joinpath("package.json").copy(src_path)
    test_path.joinpath("pub.js").copy(src_path)
    tbump.git.run_git(src_path, "init")
    tbump.git.run_git(src_path, "add", ".")
    tbump.git.run_git(src_path, "commit", "--message", "initial commit")
    return src_path


def setup_remote(tmp_path):
    git_path = tmp_path.joinpath("git")
    git_path.mkdir()
    remote_path = git_path.joinpath("repo.git")
    remote_path.mkdir()
    tbump.git.run_git(remote_path, "init", "--bare")

    src_path = tmp_path.joinpath("src")
    tbump.git.run_git(src_path, "remote", "add", "origin", remote_path)
    tbump.git.run_git(src_path, "push", "-u", "origin", "master")
    return src_path


@pytest.fixture
def test_repo(tmp_path, test_path):
    res = setup_repo(tmp_path, test_path)
    setup_remote(tmp_path)
    return res



import os
import re

import attr
import schema
import toml
import ui


@attr.s
class Config:
    current_version = attr.ib()
    version_regex = attr.ib()
    tag_template = attr.ib(default=None)
    message_template = attr.ib(default=None)
    files = attr.ib(default=None)


@attr.s
class File:
    src = attr.ib()
    search = attr.ib(default=None)
    version_template = attr.ib(default=None)


class ValidTemplate():
    def __init__(self, name, pattern):
        self.name = name
        self.pattern = pattern
        self.message = "%s should contain the string %s" % (name, pattern)

    def validate(self, value):
        if self.pattern not in value:
            raise schema.SchemaError(self.message)
        return value


class ValidTag(ValidTemplate):
    def __init__(self):
        super().__init__("tag_template", "{new_version}")


class ValidMessage(ValidTemplate):
    def __init__(self):
        super().__init__("message_template", "{new_version}")


def validate_version_template(src, version_template, known_groups):
    try:
        version_template.format(**known_groups)
    except KeyError as e:
        message = "version template for '%s' contains unknown group: %s" % (src, e)
        raise schema.SchemaError(message)


def validate(config):
    file_schema = schema.Schema({
        "src": str,
        schema.Optional("search"): str,
        schema.Optional("version_template"): str,
    })

    def compile_re(regex):
        return re.compile(regex, re.VERBOSE)

    tbump_schema = schema.Schema(
      {
        "version":  {
          "current": str,
          "regex": schema.Use(compile_re),
        },
        "git": {
          "message_template": ValidMessage(),
          "tag_template": ValidTag(),
        },
        "file": [file_schema],
        }
    )
    return tbump_schema.validate(config)


def parse(cfg_path):
    parsed = None
    with cfg_path.open() as stream:
        parsed = toml.load(stream)

    parsed = validate(parsed)
    current_version = parsed["version"]["current"]
    version_regex = parsed["version"]["regex"]
    match = version_regex.fullmatch(current_version)
    if not match:
        message = "Current version: %s does not match version regex" % current_version
        raise schema.SchemaError(message)
    current_groups = match.groupdict()
    config = Config(
        current_version=current_version,
        version_regex=version_regex
    )

    config.tag_template = parsed["git"]["tag_template"]
    config.message_template = parsed["git"]["message_template"]

    config.files = list()
    for file_dict in parsed["file"]:
        file_config = File(
            src=file_dict["src"],
            search=file_dict.get("search"),
            version_template=file_dict.get("version_template"),
        )
        if file_config.version_template:
            validate_version_template(
                file_config.src,
                file_config.version_template,
                current_groups
            )
        config.files.append(file_config)
    return config



import tbump.config
import tbump.git
import tbump.git_bumper


def test_git_bumper(test_repo):
    config = tbump.config.parse(test_repo.joinpath("tbump.toml"))
    git_bumper = tbump.git_bumper.GitBumper(test_repo)
    git_bumper.set_config(config)
    git_bumper.check_state("1.2.42")
    test_repo.joinpath("VERSION").write_text("1.2.42")
    git_bumper.bump("1.2.42")
    rc, out = tbump.git.run_git(test_repo, "log", "--oneline", raises=False)
    assert rc == 0
    assert "Bump to 1.2.42" in out



import argparse
import contextlib
import os
import sys

import path
import schema
import ui

import tbump.config
from tbump.file_bumper import FileBumper
from tbump.git_bumper import GitBumper


TBUMP_VERSION = "0.0.8"


@contextlib.contextmanager
def bump_git(git_bumper, new_version, dry_run=False):
    git_bumper.check_state(new_version)
    yield
    git_bumper.bump(new_version, dry_run=dry_run)


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("new_version")
    parser.add_argument("-C", "--cwd", dest="working_dir")
    parser.add_argument("--non-interactive", dest="interactive", action="store_false")
    parser.add_argument("--dry-run", action="store_true")
    parser.add_argument("--version", action="version", version=TBUMP_VERSION)
    args = parser.parse_args(args=args)
    interactive = args.interactive
    working_dir = args.working_dir
    new_version = args.new_version
    dry_run = args.dry_run
    if working_dir:
        os.chdir(working_dir)
    try:
        config = tbump.config.parse(path.Path("tbump.toml"))
    except IOError as io_error:
        ui.fatal("Could not read config file:", io_error)
    except Exception as e:
        ui.fatal("Invalid config:",  e)
    bumping_message = [
        "Bumping from",
        ui.reset, ui.bold, config.current_version,
        ui.reset, "to",
        ui.reset, ui.bold, new_version
    ]
    if dry_run:
        bumping_message.extend([ui.reset, ui.brown, "(dry run)"])
    ui.info_1(*bumping_message)
    working_path = path.Path.getcwd()
    git_bumper = GitBumper(working_path)
    git_bumper.set_config(config)
    file_bumper = FileBumper(working_path)
    file_bumper.set_config(config)
    with bump_git(git_bumper, new_version, dry_run=dry_run):
        changes = file_bumper.compute_changes(new_version)
        file_bumper.apply_changes(changes, dry_run=dry_run)

    if interactive and not dry_run:
        push_ok = ui.ask_yes_no("OK to push", default=False)
        if push_ok:
            git_bumper.push(new_version)



import re

import path
import pytest
import schema

import tbump.config


def test_happy_parse(monkeypatch, complex_version_regex):
    this_dir = path.Path(__file__).parent
    monkeypatch.chdir(this_dir)
    config = tbump.config.parse(path.Path("tbump.toml"))
    foo_json = tbump.config.File(
        src="package.json",
        search='"version": "{current_version}"'
    )
    version_txt = tbump.config.File(src="VERSION")
    pub_js = tbump.config.File(
        src="pub.js",
        version_template="{major}.{minor}.{patch}")

    assert config.version_regex.pattern == complex_version_regex.pattern

    assert config.files == [
        foo_json,
        version_txt,
        pub_js,
    ]

    assert config.current_version == "1.2.41-alpha-1"


def check_error(tmp_path, contents, expected_message):
    cfg_path = tmp_path.joinpath("tbump.toml")
    cfg_path.write_text(contents)
    try:
        tbump.config.parse(cfg_path)
        assert False, "shoud have raise schema error"
    except schema.SchemaError as error:
        assert expected_message in error.args[0]


def test_invalid_commit_message(tmp_path):
    check_error(
        tmp_path,
        """
        [version]
        current = '1.2'
        regex = ".*"

        [git]
        message_template = "invalid message"
        tag_template = "v{new_version}"

        [[file]]
        src = "VERSION"
        """,
        "message_template should contain the string {new_version}"
    )


def test_current_version_does_not_match_expected_regex(tmp_path):
    check_error(
        tmp_path,
        """
        [version]
        current = '1.42a1'
        regex = '(\d+)\.(\d+)\.(\d+)'

        [git]
        message_template = "Bump to  {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "VERSION"
        """,
        "Current version: 1.42a1 does not match version regex"
    )


def test_invalid_regex(tmp_path):
    check_error(
        tmp_path,
        """
        [version]
        current = '1.42a1'
        regex = '(unbalanced'

        [git]
        message_template = "Bump to  {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "VERSION"
        """,
        "Key 'regex' error"
    )


def test_invalid_custom_template(tmp_path):
    check_error(
        tmp_path,
        """
        [version]
        current = "1.2.3"
        regex = '(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)'

        [git]
        message_template = "Bump to  {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "pub.js"
        version_template = "{major}.{minor}.{no_such_group}"
        """,
        "version template for 'pub.js' contains unknown group: 'no_such_group'"
    )



import collections
import sys

import attr
import path
import ui

import tbump.config
import tbump.git


@attr.s
class Change:
    src = attr.ib()
    old = attr.ib()
    new = attr.ib()
    search = attr.ib(default=None)


@attr.s
class Replacement:
    old = attr.ib()
    new = attr.ib()


def should_replace(line, old_string, search=None):
    if not search:
        return old_string in line
    else:
        return (old_string in line) and (search in line)


def on_version_containing_none(src, verb, version, *, groups, template):
    message = [
        " ", src + ":",
        " refusing to ", verb, " version containing 'None'\n",
    ]
    message += [
        "More info:\n",
        " * version groups:  ", groups, "\n"
        " * template:        ", template, "\n",
        " * version:         ", version, "\n",
    ]
    ui.fatal(*message, end="", sep="")


def find_replacements(file_path, old_string, new_string, search=None):
    old_lines = file_path.lines()
    replacements = dict()
    for i, old_line in enumerate(old_lines):
        if should_replace(old_line, old_string, search):
            new_line = old_line.replace(old_string, new_string)
            replacements[i] = Replacement(old_line, new_line)
    return replacements


class FileBumper():
    def __init__(self, working_path):
        self.working_path = working_path
        self.files = None
        self.version_regex = None
        self.current_version = None
        self.current_groups = None
        self.new_version = None
        self.new_groups = None

    def replace_in_file(self, file_path, replacements, dry_run=False):
        self.display_replacements(file_path, replacements, dry_run=dry_run)
        if dry_run:
            return
        new_contents = ""
        old_lines = file_path.lines()
        for i, old_line in enumerate(old_lines):
            replacement = replacements.get(i)
            if replacement:
                new_contents += replacement.new
            else:
                new_contents += old_line
        file_path.write_text(new_contents)

    def display_replacements(self, file_path, replacements, dry_run=False):
        relpath = self.working_path.relpathto(file_path)
        if dry_run:
            ui.info_2("Would patch",
                      ui.reset, ui.bold, relpath)
        else:
            ui.info_2("Patching",
                      ui.reset, ui.bold, relpath)
        changed = False
        for replacement in replacements.values():
            if replacement.old != replacement.new:
                changed = True
                ui.info(ui.red, "-", replacement.old, end="")
                ui.info(ui.green, "+", replacement.new, end="")
        if not changed:
            ui.info(ui.brown, "No changes")

    def parse_version(self, version):
        match = self.version_regex.fullmatch(version)
        if match is None:
            ui.fatal("Could not parse", version, "as a valid version string")
        return match.groupdict()

    def set_config(self, config):
        self.files = config.files
        self.check_files_exist()
        self.version_regex = config.version_regex
        self.current_version = config.current_version
        self.current_groups = self.parse_version(self.current_version)

    def check_files_exist(self):
        for file in self.files:
            expected_path = self.working_path.joinpath(file.src)
            if not expected_path.exists():
                ui.fatal(file.src, "does not exist")

    def compute_changes(self, new_version):
        self.new_version = new_version
        self.new_groups = self.parse_version(self.new_version)
        res = list()
        tbump_toml_change = Change("tbump.toml", self.current_version, new_version)
        res.append(tbump_toml_change)
        for file in self.files:
            change = self.compute_changes_for_file(file)
            res.append(change)
        return res

    def compute_changes_for_file(self, file):
        if file.version_template:
            current_version = file.version_template.format(**self.current_groups)
            if "None" in current_version:
                on_version_containing_none(
                    file.src,
                    "look for",
                    current_version,
                    groups=self.current_groups,
                    template=file.version_template
                )
            new_version = file.version_template.format(**self.new_groups)
            if "None" in new_version:
                on_version_containing_none(
                    file.src,
                    "replace by",
                    new_version,
                    groups=self.new_groups,
                    template=file.version_template
                )
        else:
            current_version = self.current_version
            new_version = self.new_version

        to_search = None
        if file.search:
            to_search = file.search.format(current_version=current_version)

        return Change(file.src, current_version, new_version, search=to_search)

    def apply_changes(self, changes, dry_run=False):
        todo = collections.OrderedDict()
        errors = list()
        for change in changes:
            file_path = path.Path(self.working_path).joinpath(change.src)
            replacements = find_replacements(file_path, change.old, change.new,
                                             search=change.search)
            todo[file_path] = replacements
            if not replacements:
                errors.append(change.src)

        if errors:
            message = [" Some files did not match the old version string\n"]
            for error in errors:
                message.extend([ui.reset, " * ", ui.bold, error, "\n"])
            ui.fatal(*message, sep="", end="")

        for file_path, replacements in todo.items():
            self.replace_in_file(file_path, replacements, dry_run=dry_run)



import sys

import ui

import tbump.git


class GitBumper():
    def __init__(self, repo_path):
        self.repo_path = repo_path
        self.tag_template = None
        self.message_template = None
        self.remote_name = None
        self.remote_branch = None

    def set_config(self, config):
        self.tag_template = config.tag_template
        self.message_template = config.message_template

    def run_git(self, *args, **kwargs):
        full_args = [self.repo_path] + list(args)
        return tbump.git.run_git(*full_args, **kwargs)

    def check_dirty(self):
        rc, out = self.run_git("status", "--porcelain", raises=False)
        if rc != 0:
            ui.fatal("git status failed")
        dirty = False
        for line in out.splitlines():
            # Ignore untracked files
            if not line.startswith("??"):
                dirty = True
        if dirty:
            ui.error("Repository is dirty")
            ui.info(out)
            sys.exit(1)

    def get_current_branch(self):
        cmd = ("rev-parse", "--abbrev-ref", "HEAD")
        rc, out = self.run_git(*cmd, raises=False)
        if rc != 0:
            ui.fatal("Failed to get current ref")
        if out == "HEAD":
            ui.fatal("Not on any branch")
        return out

    def get_tracking_ref(self):
        rc, out = self.run_git(
            "rev-parse", "--abbrev-ref", "--symbolic-full-name", "@{u}",
            raises=False)
        if rc != 0:
            ui.fatal("Failed to get tracking ref")
        return out

    def check_ref_does_not_exists(self, tag_name):
        rc, _ = self.run_git("rev-parse", tag_name, raises=False)
        if rc == 0:
            ui.fatal("git ref", tag_name, "already exists")

    def check_state(self, new_version):
        self.check_dirty()
        branch_name = self.get_current_branch()
        tracking_ref = self.get_tracking_ref()
        self.remote_name, self.remote_branch = tracking_ref.split("/", maxsplit=1)

        tag_name = self.tag_template.format(new_version=new_version)
        self.check_ref_does_not_exists(tag_name)

    def commit(self, message, dry_run=False):
        if dry_run:
            ui.info_2("Would commit with message: '%s'" % message)
        else:
            ui.info_2("Making bump commit")
            self.run_git("add", "--update")
            self.run_git("commit", "--message", message)

    def tag(self, tag_name, dry_run=False):
        if dry_run:
            ui.info_2("Would create tag:", tag_name)
        else:
            ui.info_2("Creating tag", tag_name)
            self.run_git("tag", tag_name)

    def bump(self, new_version, dry_run=False):
        message = self.message_template.format(new_version=new_version)
        self.commit(message, dry_run=dry_run)
        tag_name = self.tag_template.format(new_version=new_version)
        self.tag(tag_name, dry_run=dry_run)

    def push(self, new_version):
        tag_name = self.tag_template.format(new_version=new_version)
        self.run_git("push", self.remote_name, self.remote_branch, tag_name, verbose=True)



import pytest

import tbump.file_bumper
from tbump.test.conftest import assert_in_file
from ui.tests.conftest import message_recorder


def test_file_bumper(test_repo):
    bumper = tbump.file_bumper.FileBumper(test_repo)
    config = tbump.config.parse(test_repo.joinpath("tbump.toml"))
    assert bumper.working_path == test_repo
    bumper.set_config(config)
    changes = bumper.compute_changes(new_version="1.2.41-alpha-2")
    assert changes == [
        tbump.file_bumper.Change("tbump.toml", "1.2.41-alpha-1", "1.2.41-alpha-2"),
        tbump.file_bumper.Change(
            "package.json", "1.2.41-alpha-1", "1.2.41-alpha-2",
            search='"version": "1.2.41-alpha-1"',
        ),
        tbump.file_bumper.Change("VERSION", "1.2.41-alpha-1", "1.2.41-alpha-2"),
        tbump.file_bumper.Change("pub.js", "1.2.41", "1.2.41"),
    ]

    bumper.apply_changes(changes)

    assert_in_file(test_repo, "package.json", '"version": "1.2.41-alpha-2"')
    assert_in_file(test_repo, "package.json", '"other-dep": "1.2.41-alpha-1"')
    assert_in_file(test_repo, "pub.js", "PUBLIC_VERSION = '1.2.41'")


def test_looking_for_empty_groups(tmp_path, message_recorder):
    tbump_path = tmp_path.joinpath("tbump.toml")
    tbump_path.write_text(
        """
        [version]
        current = "1.2"
        regex = '''
            (?P<major>\d+)
            \.
            (?P<minor>\d+)
            (
              \.
              (?P<patch>\d+)
            )?
        '''

        [git]
        message_template = "Bump to {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "foo"
        version_template = "{major}.{minor}.{patch}"

        """
    )
    foo_path = tmp_path.joinpath("foo")
    foo_path.write_text(
        """
        version = "1.2"
        """
    )
    config = tbump.config.parse(tbump_path)
    bumper = tbump.file_bumper.FileBumper(tmp_path)
    bumper.set_config(config)
    with pytest.raises(SystemExit) as e:
        bumper.compute_changes(new_version="1.3.1")
    assert message_recorder.find("refusing to look for version containing 'None'")


def test_replacing_with_empty_groups(tmp_path, message_recorder):
    tbump_path = tmp_path.joinpath("tbump.toml")
    tbump_path.write_text(
        """
        [version]
        current = "1.2.3"
        regex = '''
            (?P<major>\d+)
            \.
            (?P<minor>\d+)
            (
              \.
              (?P<patch>\d+)
            )?
        '''

        [git]
        message_template = "Bump to {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "foo"
        version_template = "{major}.{minor}.{patch}"

        """
    )
    foo_path = tmp_path.joinpath("foo")
    foo_path.write_text(
        """
        version = "1.2.3"
        """
    )

    bumper = tbump.file_bumper.FileBumper(tmp_path)
    config = tbump.config.parse(tbump_path)
    bumper.set_config(config)
    with pytest.raises(SystemExit):
        changes = bumper.compute_changes(new_version="1.3")

    assert message_recorder.find("refusing to replace by version containing 'None'")



import subprocess
import textwrap

import toml
import path
import pytest
from ui.tests.conftest import message_recorder
from tbump.test.conftest import assert_in_file

import tbump.main
import tbump.git


def test_replaces(test_repo):
    tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])

    toml_path = test_repo.joinpath("tbump.toml")
    new_toml = toml.loads(toml_path.text())
    assert new_toml["version"]["current"] == "1.2.41-alpha-2"

    assert_in_file("package.json", '"version": "1.2.41-alpha-2"')
    assert_in_file("package.json", '"other-dep": "1.2.41-alpha-1"')
    assert_in_file("pub.js", "PUBLIC_VERSION = '1.2.41'")


def test_new_version_does_not_match(test_repo, message_recorder):
    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41a2", "--non-interactive"])
    assert message_recorder.find("Could not parse 1.2.41a2")


def test_abort_if_file_does_not_exist(test_repo, message_recorder):
    test_repo.joinpath("package.json").remove()
    tbump.git.run_git(test_repo, "add", "--update")
    tbump.git.run_git(test_repo, "commit", "--message", "remove package.json")
    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])
    assert message_recorder.find("package.json does not exist")


def test_commit_and_tag(test_repo):
    tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])

    rc, out = tbump.git.run_git(test_repo, "log", "--oneline", raises=False)
    assert rc == 0
    assert "Bump to 1.2.41-alpha-2" in out

    rc, out = tbump.git.run_git(test_repo, "tag", "--list", raises=False)
    assert rc == 0
    assert out == "v1.2.41-alpha-2"


def test_abort_if_dirty(test_repo, message_recorder):
    test_repo.joinpath("VERSION").write_text("unstaged changes\n", append=True)

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])
    assert message_recorder.find("dirty")


def test_abort_if_tag_exists(test_repo, message_recorder):
    tbump.git.run_git(test_repo, "tag", "v1.2.42")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    assert message_recorder.find("1.2.42 already exists")


def test_abort_if_file_does_not_match(test_repo, message_recorder):
    invalid_src = test_repo.joinpath("foo.txt")
    invalid_src.write_text("this is foo")
    tbump_path = test_repo.joinpath("tbump.toml")
    tbump_path.write_text("""\
    [[file]]
    src = "foo.txt"
    """, append=True)
    tbump.git.run_git(test_repo, "add", ".")
    tbump.git.run_git(test_repo, "commit", "--message", "add foo.txt")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    assert message_recorder.find("did not match")
    assert message_recorder.find("foo\.txt")
    assert_in_file("VERSION", "1.2.41-alpha-1")


def test_interactive_push(test_repo, message_recorder, mock):
    ask_mock = mock.patch("ui.ask_yes_no")
    ask_mock.return_value = True
    tbump.main.main(["-C", test_repo, "1.2.42"])
    ask_mock.assert_called_with("OK to push", default=False)
    rc, out = tbump.git.run_git(test_repo, "ls-remote", raises=False)
    assert rc == 0
    assert "tags/v1.2.42" in out


def test_do_not_add_untracked_files(test_repo):
    test_repo.joinpath("untracked.txt").write_text("please don't add me")
    tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    rc, out = tbump.git.run_git(test_repo, "show", "--stat", "HEAD", raises=False)
    assert rc == 0
    assert "untracked.txt" not in out


def test_dry_run(test_repo):
    tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--dry-run"])

    toml_path = test_repo.joinpath("tbump.toml")
    assert_in_file("VERSION", "1.2.41-alpha-1")
